from packaging import version
from importlib_metadata import version as lib_version
import requests, shutil, sys, subprocess, re


def isCommandInstalled(package):
    return True if shutil.which(package.split(" ", 1)[0]) else False

def getPackageInstalledVersion(package):
    try:
        installed_version = lib_version(package)
        return version.parse(installed_version)
    except:
        return None

def getPackageLatestVersion(package):
    try:
        response = requests.get(f"https://pypi.org/pypi/{package}/json", timeout=10)
        latest_version = response.json()['info']['version']
        return version.parse(latest_version)
    except:
        return None

def list_installed_packages():
    import importlib.metadata
    packages = importlib.metadata.distributions()
    for package in packages:
        print(f"{package.metadata['Name']} {package.version}")

def updatePip():
    if isCommandInstalled("pip"):
        pipInstallCommand = f"{sys.executable} -m pip install"
        pipFailedUpdated = "pip tool failed to be updated!"
        try:
            # Update pip tool in case it is too old
            updatePip = subprocess.Popen(f"{pipInstallCommand} --upgrade pip", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            *_, stderr = updatePip.communicate()
            if not stderr:
                print("pip tool updated!")
            else:
                print(pipFailedUpdated)
        except:
            print(pipFailedUpdated)

def installPipPackage(module, update=False):
    #executablePath = os.path.dirname(sys.executable)
    #pippath = os.path.join(executablePath, "pip")
    #pip = pippath if os.path.isfile(pippath) else "pip"
    #pip3path = os.path.join(executablePath, "pip3")
    #pip3 = pip3path if os.path.isfile(pip3path) else "pip3"

    if isCommandInstalled("pip"):
        pipInstallCommand = f"{sys.executable} -m pip install"

        if update:
            updatePip()
        try:
            upgrade = (module.startswith("-U ") or module.startswith("--upgrade "))
            if upgrade:
                moduleName = re.sub("^[^ ]+? (.+?)$", r"\1", module)
            else:
                moduleName = module
            print(f"{'Upgrading' if upgrade else 'Installing'} '{moduleName}' ...")
            installNewModule = subprocess.Popen(f"{pipInstallCommand} {module}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            *_, stderr = installNewModule.communicate()
            if not stderr:
                print(f"Package '{moduleName}' {'upgraded' if upgrade else 'installed'}!")
            else:
                print(f"Failed {'upgrading' if upgrade else 'installing'} package '{moduleName}'!")
                print(stderr)
            return True
        except:
            return False

    else:
        print("pip command is not found!")
        return False


from toolmate_sdk.utils.online import openURL
import urllib.parse

TOOL_SYSTEM = "You are an expert in writing emails, specializing in generating email title and content based on user requests. You are also skilled at finding the recipient's email address from the information provided."

TOOL_SCHEMA = {
    "name": "send_gmail",
    "description": "Send Gmail",
    "parameters": {
        "type": "object",
        "properties": {
            "email_address": {
                "type": "string",
                "description": "The recipient of the email.",
            },
            "email_title": {
                "type": "string",
                "description": "Give a title to the email.",
            },
            "eamil_content": {
                "type": "string",
                "description": "The body or content of the email.",
            },
        },
        "required": ["email_title", "eamil_content"],
    },
}

def send_gmail(email_title: str, eamil_content: str, email_address: str=""):

    subject = urllib.parse.quote(email_title)
    body = urllib.parse.quote(eamil_content)

    def getGoogleLink():
        link = "https://mail.google.com/mail/?view=cm&fs=1"
        if email_address:
            link += f"&to={email_address}"
        if subject:
            link += f"&su={subject}"
        if body:
            link += f"&body={body}"
        return link

    openURL(getGoogleLink())
    return ""

TOOL_METHOD = send_gmail


from .text_wrapper import TextWrapper
from typing import Optional, Any
import threading, traceback, shutil, textwrap

def wrapText(content, terminal_width=None):
    if terminal_width is None:
        terminal_width = shutil.get_terminal_size().columns
    return "\n".join([textwrap.fill(line, width=terminal_width) for line in content.split("\n")])

def getChatCompletionText(
        backend: str,
        completion: Any,
        stream: Optional[bool]=False,
        print_on_terminal: Optional[bool]=True,
        word_wrap: Optional[bool]=True,
    ) -> str:
    if stream:
        text_output = readStreamingChunks(backend, completion, print_on_terminal, word_wrap)
        # TODO: ensure client connection, e.g. llama.cpp client, is closed properly
    else:
        if backend == "anthropic":
            text_output = completion.content[0].text
        elif backend == "ollama":
            text_output = completion.message.content
        elif backend in ("genai", "vertexai"):
            text_output = completion.candidates[0].content.parts[0].text
        elif backend in ("azure", "custom", "deepseek", "github", "googleai", "groq", "llamacpp", "mistral", "openai", "xai"):
            text_output = completion.choices[0].message.content
        if print_on_terminal:
            print(wrapText(text_output) if word_wrap else text_output)
    return text_output

def readStreamingChunks(
        backend: str,
        completion: Any,
        print_on_terminal: Optional[bool]=True,
        word_wrap: Optional[bool]=True,
    ) -> str:
    openai_style = True if backend in ("azure", "custom", "deepseek", "github", "googleai", "groq", "llamacpp", "mistral", "openai", "xai") else False
    try:
        text_wrapper = TextWrapper(word_wrap)
        streaming_event = threading.Event()
        streaming_thread = threading.Thread(target=text_wrapper.streamOutputs, args=(streaming_event, completion, openai_style, print_on_terminal))
        # Start the streaming thread
        streaming_thread.start()
        # wait while text output is steaming; capture key combo 'ctrl+q' or 'ctrl+z' to stop the streaming
        text_wrapper.keyToStopStreaming(streaming_event)
        # when streaming is done or when user press "ctrl+q"
        streaming_thread.join()
    except:
        print(traceback.format_exc())
        return "" # Incomplete streaming
    return text_wrapper.text_output # completion streaming is successful


from .backends.anthropic import AnthropicLLM
from .backends.azure import AzureLLM
from .backends.custom import OpenaiCompatibleLLM
from .backends.deepseek import DeepseekLLM
from .backends.genai import GenaiLLM
from .backends.github import GithubLLM
from .backends.googleai import GoogleaiLLM
from .backends.groq import GroqLLM
from .backends.llamacpp import LlamacppLLM
from .backends.mistral import MistralLLM
from .backends.ollama import OllamaLLM
from .backends.openai import OpenaiLLM
from .backends.xai import XaiLLM

from .utils.instructions import getRagPrompt
from .utils.retrieve_text_output import getChatCompletionText
from .utils.handle_text import readTextFile

from typing import Optional, Callable, Union, Any, List, Dict
from copy import deepcopy
from io import StringIO
import sys, os, json, traceback

PACKAGE_PATH = os.path.dirname(os.path.realpath(__file__))
PACKAGE_NAME = os.path.basename(PACKAGE_PATH)
SUPPORTED_BACKENDS = ["anthropic", "azure", "custom", "deepseek", "genai", "github", "googleai", "groq", "llamacpp", "mistral", "ollama", "openai", "vertexai", "xai"]
DEVELOPER_MODE = True if os.getenv("DEVELOPER_MODE") == "TRUE" else False

def generate(
        messages: Union[List[Dict[str, str]], str], # user request or messages containing user request; accepts either a single string or a list of dictionaries
        backend: Optional[str]="ollama", # AI backend; check SUPPORTED_BACKENDS for supported backends
        model: Optional[str]=None, # AI model name; applicable to all backends, execept for llamacpp
        model_keep_alive: Optional[str]=None, # time to keep the model loaded in memory; applicable to ollama only
        system: Optional[Union[List[str], str]]=None, # system message; define how the model should generally behave and respond; accepts a list of strings or a single string; loop through multiple system messages for multi-turn inferences if it is a list
        context: Optional[Union[List[str], str]]=None, # predefined context to be added to the user prompt as prefix; accepts a list of strings or a single string; loop through multiple predefined contexts for multi-turn inferences if it is a list
        follow_up_prompt: Optional[Union[List[str], str]]=None, # follow-up prompts after an assistant message is generated; accepts a list of strings or a single string; loop through multiple follow-up prompts for multi-turn inferences if it is a list
        tool: Optional[Union[List[str], str]]=None, # a tool either a built-in tool name under the folder `tools` in the package directory or a file path of the tool; accepts a list of strings or a single string; loop through multiple tools for multi-turn actions if it is a list; parameters of both `schema` and `func` are ignored when `tool` parameter is given
        schema: Optional[dict]=None, # json schema for structured output or function calling
        func: Optional[Callable[..., Optional[str]]]=None, # function to be called
        temperature: Optional[float]=None, # temperature for sampling
        max_tokens: Optional[int]=None, # maximum number of tokens to generate
        context_window: Optional[int]=None, # context window size; applicable to ollama only
        batch_size: Optional[int]=None, # batch size; applicable to ollama only
        prefill: Optional[Union[List[str], str]]=None, # prefill of assistant message; applicable to deepseek, mistral, ollama and groq only; accepts a list of strings or a single string; loop through multiple prefills for multi-turn inferences if it is a list
        stop: Optional[list]=None, # stop sequences
        stream: Optional[bool]=False, # stream partial message deltas as they are available
        stream_events_only: Optional[bool]=False, # return streaming events object only
        api_key: Optional[str]=None, # API key or credentials json file path in case of using Vertex AI as backend; applicable to anthropic, custom, deepseek, genai, github, googleai, groq, mistral, openai, xai
        api_endpoint: Optional[str]=None, # API endpoint; applicable to anthropic, azure, custom, llamacpp, ollama
        api_project_id: Optional[str]=None, # project id; applicable to Vertex AI only, i.e., vertexai or genai
        api_service_location: Optional[str]=None, # cloud service location; applicable to Vertex AI only, i.e., vertexai or genai
        api_timeout: Optional[Union[int, float]]=None, # timeout for API request; applicable to backends, execept for ollama, genai and vertexai
        print_on_terminal: Optional[bool]=True, # print output on terminal
        word_wrap: Optional[bool]=True, # word wrap output according to current terminal width
        **kwargs, # pass extra options supported by individual backends
) -> Union[List[Dict[str, str]], Any]:
    """
    Generate AI assistant response.

    Args:
        messages:
            type: Union[List[Dict[str, str]], str]
            user request or messages containing user request
            accepts either a single string or a list of dictionaries
            use a single string string to specify user request without chat history
            use a list of dictionaries to provide with the onging interaction between user and assistant
            when a list is given:
                each dictionary in the list should contain keys `role` and `content`
                specify the latest user request in the last item content
                list format example:
                    [
                        {"role": "system", "You are an AI assistant."},
                        {"role": "user", "Hello!"},
                        {"role": "assistant", "Hello! How can I assist you today?"},
                        {"role": "user", "What is generative AI?"}
                    ]

        backend:
            type: Optional[str]="ollama"
            AI backend
            supported backends: "anthropic", "azure", "custom", "deepseek", "genai", "github", "googleai", "groq", "llamacpp", "mistral", "ollama", "openai", "vertexai", "xai"

        model:
            type: Optional[str]=None
            AI model name
            applicable to all backends, execept for `llamacpp`
            for backend `llamacpp`, specify a model file in the command line running the llama.cpp server
            for backend `ollama`, model is automatically downloaded if it is not in the downloaded model list

        model_keep_alive:
            type: Optional[str]=None
            time to keep the model loaded in memory
            applicable to ollama only

        system:
            type: Optional[Union[List[str], str]]=None
            system message
            define how the model should generally behave and respond
            accepts a list of strings or a single string
            loop through multiple system messages for multi-turn inferences if it is a list

        context:
            type: Optional[Union[List[str], str]]=None
            predefined context to be added to the user prompt as prefix
            accepts a list of strings or a single string
            loop through multiple predefined contexts for multi-turn inferences if it is a list

        follow_up_prompt:
            type: Optional[Union[List[str], str]]=None
            follow-up prompts after an assistant message is generated
            accepts a list of strings or a single string
            loop through multiple follow-up prompts for multi-turn inferences if it is a list

        tool:
            type: Optional[Union[List[str], str]]=None
            a tool either a built-in tool name under the folder `tools` in the package directory or a file path of the tool
            accepts a list of strings or a single string
            loop through multiple tools for multi-turn actions if it is a list
            parameters of both `schema` and `func` are ignored when `tool` parameter is given

        schema:
            type: Optional[dict]=None
            json schema for structured output or function calling

        func:
            type: Optional[Callable[..., Optional[str]]]=None
            function to be called

        temperature:
            type: Optional[float]=None
            temperature for sampling

        max_tokens:
            type: Optional[int]=None
            maximum number of tokens to generate

        context_window:
            type: Optional[int]=None
            context window size
            applicable to ollama only

        batch_size:
            type: Optional[int]=None
            batch size
            applicable to ollama only

        prefill:
            type: Optional[Union[List[str], str]]=None
            prefill of assistant message
            applicable to deepseek, mistral, ollama and groq only
            accepts a list of strings or a single string
            loop through multiple prefills for multi-turn inferences if it is a list

        stop:
            type: Optional[list]=None
            stop sequences

        stream:
            type: Optional[bool]=False
            stream partial message deltas as they are available

        stream_events_only:
            type: Optional[bool]=False
            return streaming events object only

        api_key:
            type: Optional[str]=None
            API key or credentials json file path in case of using Vertex AI as backend
            applicable to anthropic, custom, deepseek, genai, github, googleai, groq, mistral, openai, xai

        api_endpoint:
            type: Optional[str]=None
            API endpoint
            applicable to anthropic, azure, custom, llamacpp, ollama

        api_project_id:
            type: Optional[str]=None
            project id
            applicable to Vertex AI only, i.e., vertexai or genai

        api_service_location:
            type: Optional[str]=None
            cloud service location
            applicable to Vertex AI only, i.e., vertexai or genai

        api_timeout:
            type: Optional[Union[int, float]]=None
            timeout for API request
            applicable to backends, execept for ollama, genai and vertexai

        print_on_terminal:
            type: Optional[bool]=True
            print output on terminal

        word_wrap:
            type: Optional[bool]=True
            word wrap output according to current terminal width

        **kwargs,
            pass extra options supported by individual backends

    Return:
        either:
            list of messages containing multi-turn interaction between user and the AI assistant
            find the latest assistant response in the last item of the list
        or:
            streaming events object of AI assistant response when both parameters `stream` and `stream_events_only` are set to `True`
    """
    if backend not in SUPPORTED_BACKENDS:
        raise ValueError(f"Backend {backend} is not supported. Supported backends are {SUPPORTED_BACKENDS}")
    # placeholders
    original_system = ""
    chat_system = ""
    # deep copy messages avoid modifying the original one
    messagesCopy = deepcopy(messages) if isinstance(messages, list) else [{"role": "system", "content": "You are a helpful AI assistant."}, {"role": "user", "content": messages}]
    # handle given system message(s)
    if system:
        if isinstance(system, list):
            system_instruction = system.pop(0)
        else: # a string instead
            system_instruction = system
            system = []
        # check if it is a predefined system message built-in with this SDK
        possible_system_file_path = os.path.join(PACKAGE_PATH, "systems", f"{system_instruction}.md")
        if os.path.isfile(possible_system_file_path):
            system_file_content = readTextFile(possible_system_file_path)
            if system_file_content:
                system_instruction = system_file_content
        elif os.path.isfile(system_instruction): # system_instruction itself is a valid filepath
            system_file_content = readTextFile(system_instruction)
            if system_file_content:
                system_instruction = system_file_content
        original_system = updateSystemMessage(messagesCopy, system_instruction)
    # handle given predefined context(s)
    if context:
        if isinstance(context, list):
            context_content = context.pop(0)
        else: # a string instead
            context_content = context
            context = []
        # check if it is a predefined context built-in with this SDK
        possible_context_file_path = os.path.join(PACKAGE_PATH, "contexts", f"{context_content}.md")
        if os.path.isfile(possible_context_file_path):
            context_file_content = readTextFile(possible_context_file_path)
            if context_file_content:
                context_content = context_file_content
        elif os.path.isfile(context_content): # context_content itself is a valid filepath
            context_file_content = readTextFile(context_content)
            if context_file_content:
                context_content = context_file_content
        messagesCopy[-1] = context_content + messagesCopy[-1]
    # handle given prefill(s)
    if prefill:
        if isinstance(prefill, list):
            prefill_content = prefill.pop(0)
        else: # a string instead
            prefill_content = prefill
            prefill = None
    else:
        prefill_content = None
    # handle given tools
    if tool:
        if isinstance(tool, list):
            tool_entity = tool.pop(0)
        else: # a string instead
            tool_entity = tool
            tool = []
        tool_name = tool_entity[:20]
        # check if it is a predefined tool built-in with this SDK
        possible_tool_file_path = os.path.join(PACKAGE_PATH, "tools", f"{tool_entity}.py")
        if os.path.isfile(possible_tool_file_path):
            tool_file_content = readTextFile(possible_tool_file_path)
            if tool_file_content:
                tool_entity = tool_file_content
        elif os.path.isfile(tool_entity): # tool_entity itself is a valid filepath
            tool_file_content = readTextFile(tool_entity)
            if tool_file_content:
                tool_entity = tool_file_content
        try:
            exec(tool_entity, globals())
            tool_system = TOOL_SYSTEM
            schema = TOOL_SCHEMA
            func = TOOL_METHOD
            if tool_system:
                chat_system = updateSystemMessage(messagesCopy, tool_system)
        except Exception as e:
            print(f"Failed to execute tool `{tool_name}`! An error occurred: {e}")
            if DEVELOPER_MODE:
                print(traceback.format_exc())

    # check if it is last request
    is_last_request = True if not follow_up_prompt and not system and not context and not tool and not prefill else False
    
    # deep copy schema avoid modifying the original one
    schemaCopy = None if schema is None else deepcopy(schema)
    # run LLM
    if schemaCopy is not None: # structured output or function calling; allow schema to be an empty dict
        dictionary_output = {} if not schemaCopy else getDictionaryOutput(
            messagesCopy,
            schemaCopy,
            backend,
            model=model,
            model_keep_alive=model_keep_alive,
            temperature=temperature,
            max_tokens=max_tokens,
            context_window=context_window,
            batch_size=batch_size,
            prefill=prefill_content,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            api_timeout=api_timeout,
            **kwargs
        )
        if chat_system:
            updateSystemMessage(messagesCopy, chat_system)
            chat_system = ""
        if func:
            # Create a StringIO object to capture the output
            terminal_output = StringIO()
            # Redirect stdout to the StringIO object
            old_stdout = sys.stdout
            sys.stdout = terminal_output
            # placeholder for function text output
            function_text_output = ""
            try:
                # execute the function
                function_response = func() if not dictionary_output else func(**dictionary_output) # returned response can be either 1) an empty string: no chat extension 2) a non-empty string: chat extension 3) none: errors encountered in executing the function
                function_text_output = terminal_output.getvalue() # capture the function text output for function calling without chat extension
            except:
                function_response = None # due to unexpected errors encountered in executing the function; fall back to regular completion
            # Restore the original stdout
            sys.stdout = old_stdout
            # handle function response
            if function_response is None or function_response: # fall back to regular completion if function_response is None; chat extension if function_response
                if function_response:
                    # added function response as context to the original prompt
                    addContextToMessages(messagesCopy, function_response)
                return generate(
                    messagesCopy,
                    backend,
                    model=model,
                    model_keep_alive=model_keep_alive,
                    system=None if function_response else system,
                    context=None if function_response else context,
                    follow_up_prompt=None if function_response else follow_up_prompt,
                    tool=None if function_response else tool,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    context_window=context_window,
                    batch_size=batch_size,
                    prefill=None if function_response else prefill_content,
                    stop=stop,
                    stream=stream,
                    stream_events_only=stream_events_only,
                    api_key=api_key,
                    api_endpoint=api_endpoint,
                    api_project_id=api_project_id,
                    api_service_location=api_service_location,
                    api_timeout=api_timeout,
                    print_on_terminal=print_on_terminal,
                    word_wrap=word_wrap,
                    **kwargs
                )
            else: # empty str; function executed successfully without chat extension
                output = function_text_output if function_text_output else "Done!"
        else: # structured output
            output = json.dumps(dictionary_output)
        if print_on_terminal:
            print(output)
    else: # regular completion
        if backend == "anthropic":
            completion = AnthropicLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "azure":
            completion = AzureLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_endpoint=api_endpoint,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "custom":
            completion = OpenaiCompatibleLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_endpoint=api_endpoint,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "deepseek":
            completion = DeepseekLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                prefill=prefill_content,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend in ("genai", "vertexai"):
            completion = GenaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_project_id=api_project_id,
                api_service_location=api_service_location,
                **kwargs
            )
        elif backend == "github":
            completion = GithubLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "googleai":
            completion = GoogleaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "groq":
            completion = GroqLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                prefill=prefill_content,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "llamacpp":
            completion = LlamacppLLM.getChatCompletion(
                messagesCopy,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_endpoint=api_endpoint,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "mistral":
            completion = MistralLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                prefill=prefill_content,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "ollama":
            completion = OllamaLLM.getChatCompletion(             
                messagesCopy,
                model=model,
                model_keep_alive=model_keep_alive,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                context_window=context_window,
                batch_size=batch_size,
                prefill=prefill_content,
                stop=stop,
                stream=stream,
                api_endpoint=api_endpoint,
                **kwargs
            )
        elif backend == "openai":
            completion = OpenaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "xai":
            completion = XaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        if stream and stream_events_only and is_last_request:
            return completion
        output = getChatCompletionText(backend, completion, stream=stream, print_on_terminal=print_on_terminal, word_wrap=word_wrap)
    
    # update the message list
    messagesCopy.append({"role": "assistant", "content": output})
    if original_system:
        updateSystemMessage(messagesCopy, original_system)
    # work on follow-up prompts
    if not is_last_request and not follow_up_prompt:
        follow_up_prompt = "Tell me more"
    if follow_up_prompt:
        if isinstance(follow_up_prompt, list):
            follow_up_prompt_content = follow_up_prompt.pop(0)
        else: # a string instead
            follow_up_prompt_content = follow_up_prompt
            follow_up_prompt = []
        # check if it is a predefined follow_up_prompt built-in with this SDK
        possible_follow_up_prompt_file_path = os.path.join(PACKAGE_PATH, "prompts", f"{follow_up_prompt_content}.md")
        if os.path.isfile(possible_follow_up_prompt_file_path):
            follow_up_prompt_file_content = readTextFile(possible_follow_up_prompt_file_path)
            if follow_up_prompt_file_content:
                follow_up_prompt_content = follow_up_prompt_file_content
        elif os.path.isfile(follow_up_prompt_content): # follow_up_prompt_content itself is a valid filepath
            follow_up_prompt_file_content = readTextFile(follow_up_prompt_content)
            if follow_up_prompt_file_content:
                follow_up_prompt_content = follow_up_prompt_file_content
        messagesCopy.append({"role": "user", "content": follow_up_prompt_content})
        return generate(
            messages=messagesCopy,
            backend=backend,
            model=model,
            model_keep_alive=model_keep_alive,
            system=system,
            context=context,
            follow_up_prompt=follow_up_prompt,
            tool=tool,
            temperature=temperature,
            max_tokens=max_tokens,
            context_window=context_window,
            batch_size=batch_size,
            prefill=prefill,
            stop=stop,
            stream=stream,
            stream_events_only=stream_events_only,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            api_timeout=api_timeout,
            print_on_terminal=print_on_terminal,
            word_wrap=word_wrap,
            **kwargs
        )
    return messagesCopy

def getDictionaryOutput(
        messages: List[Dict[str, str]],
        schema: dict,
        backend: str,
        model: Optional[str]=None,
        model_keep_alive: Optional[str]=None,
        temperature: Optional[float]=None, 
        max_tokens: Optional[int]=None,
        context_window: Optional[int]=None,
        batch_size: Optional[int]=None,
        prefill: Optional[str]=None,
        stop: Optional[list]=None,
        api_key: Optional[str]=None,
        api_endpoint: Optional[str]=None,
        api_project_id: Optional[str]=None,
        api_service_location: Optional[str]=None,
        api_timeout: Optional[Union[int, float]]=None,
        **kwargs,
) -> dict:
    """
    Returns dictionary in response to user message
    """
    if backend == "anthropic":
        return AnthropicLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "azure":
        return AzureLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "custom":
        return OpenaiCompatibleLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "deepseek":
        return DeepseekLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend in ("genai", "vertexai"):
        return GenaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            **kwargs
        )
    elif backend == "github":
        return GithubLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "googleai":
        return GoogleaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "groq":
        return GroqLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "llamacpp":
        return LlamacppLLM.getDictionaryOutput(
            messages,
            schema,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "mistral":
        return MistralLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "ollama":
        return OllamaLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            model_keep_alive=model_keep_alive,
            temperature=temperature,
            max_tokens=max_tokens,
            context_window=context_window,
            batch_size=batch_size,
            prefill=prefill,
            stop=stop,
            api_endpoint=api_endpoint,
            **kwargs
        )
    elif backend == "openai":
        return OpenaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "xai":
        return XaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    return {}

def updateSystemMessage(messages: List[Dict[str, str]], system: str) -> str:
    """
    update system message content in the given message list
    and return the original system message
    """
    original_system = ""
    for i in messages:
        if i.get("role", "") == "system":
            original_system = i.get("content", "")
            i["content"] = system
            break
    return original_system

def addContextToMessages(messages: List[Dict[str, str]], context: str):
    print(4564564, context)
    """
    add context to user prompt
    assuming user prompt is placed in the last item of the given message list
    """
    messages[-1] = {"role": "user", "content": getRagPrompt(messages[-1].get("content", ""), context)}


import os
from typing import Union

def readTextFile(textFile: str) -> Union[str, None]:
    if not os.path.isfile(textFile):
        return None
    with open(textFile, 'r', encoding='utf8') as fileObj:
        content = fileObj.read()
    return content if content else ""


from prompt_toolkit.keys import Keys
from prompt_toolkit.input import create_input
import asyncio, shutil, os, wcwidth


class TextWrapper:

    def __init__(self, wrap_words: bool=True):
        self.streaming_finished = False
        self.word_wrap = wrap_words
        self.text_chunk = ""
        self.text_output = ""

    def getStringWidth(self, text):
        width = 0
        for character in text:
            width += wcwidth.wcwidth(character)
        return width

    def wrapStreamWords(self, answer, terminal_width):
        if " " in answer:
            if answer == " ":
                if self.line_width < terminal_width:
                    print(" ", end='', flush=True)
                    self.line_width += 1
            else:
                answers = answer.split(" ")
                for index, item in enumerate(answers):
                    isLastItem = (len(answers) - index == 1)
                    itemWidth = self.getStringWidth(item)
                    newLineWidth = (self.line_width + itemWidth) if isLastItem else (self.line_width + itemWidth + 1)
                    if isLastItem:
                        if newLineWidth > terminal_width:
                            print(f"\n{item}", end='', flush=True)
                            self.line_width = itemWidth
                        else:
                            print(item, end='', flush=True)
                            self.line_width += itemWidth
                    else:
                        if (newLineWidth - terminal_width) == 1:
                            print(f"{item}\n", end='', flush=True)
                            self.line_width = 0
                        elif newLineWidth > terminal_width:
                            print(f"\n{item} ", end='', flush=True)
                            self.line_width = itemWidth + 1
                        else:
                            print(f"{item} ", end='', flush=True)
                            self.line_width += (itemWidth + 1)
        else:
            answerWidth = self.getStringWidth(answer)
            newLineWidth = self.line_width + answerWidth
            if newLineWidth > terminal_width:
                print(f"\n{answer}", end='', flush=True)
                self.line_width = answerWidth
            else:
                print(answer, end='', flush=True)
                self.line_width += answerWidth

    def keyToStopStreaming(self, streaming_event):
        async def readKeys() -> None:
            done = False
            input = create_input()

            def keys_ready():
                nonlocal done
                for key_press in input.read_keys():
                    #print(key_press)
                    if key_press.key in (Keys.ControlQ, Keys.ControlZ):
                        print("\n")
                        done = True
                        streaming_event.set()

            with input.raw_mode():
                with input.attach(keys_ready):
                    while not done:
                        if self.streaming_finished:
                            break
                        await asyncio.sleep(0.1)

        try:
            asyncio.run(readKeys())
        except:
            pass

    def streamOutputs(self, streaming_event, completion, openai_style=False, print_on_terminal=True):
        terminal_width = shutil.get_terminal_size().columns

        def finishOutputs(word_wrap, chat_response):
            self.word_wrap = word_wrap
            self.text_chunk = ""
            self.streaming_finished = True
            self.text_output = chat_response

        chat_response = ""
        self.line_width = 0
        block_start = False
        word_wrap = self.word_wrap
        first_event = True
        stop_file = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "temp", "stop_running")
        for event in completion:
            if os.path.isfile(stop_file):
                os.remove(stop_file)
                finishOutputs(word_wrap, chat_response)
            if streaming_event is None or (not streaming_event.is_set() and not self.streaming_finished):
                # RETRIEVE THE TEXT FROM THE RESPONSE
                if openai_style:
                    # openai
                    # when open api key is invalid for some reasons, event response in string
                    if isinstance(event, str):
                        answer = event
                    elif hasattr(event, "data"): # mistralai
                        answer = event.data.choices[0].delta.content
                    elif not event.choices: # in case of the 1st event of azure's completion
                        continue
                    else:
                        answer = event.choices[0].delta.content
                elif hasattr(event, "delta") and hasattr(event.delta, "text"): # anthropic
                    answer = event.delta.text
                elif hasattr(event, "content_block") and hasattr(event.content_block, "text"):
                    answer = event.content_block.text
                elif str(type(event)).startswith("<class 'anthropic.types"): # anthropic
                    continue
                elif hasattr(event, "message"): # newer ollama python package
                    answer = event.message.content
                elif isinstance(event, dict):
                    if "message" in event:
                        # ollama chat
                        answer = event["message"].get("content", "")
                    else:
                        # llama.cpp chat
                        answer = event["choices"][0]["delta"].get("content", "")
                else:
                    # vertex ai, genai
                    answer = event.text
                # STREAM THE ANSWER
                if answer is not None:
                    if first_event:
                        first_event = False
                        answer = answer.lstrip()
                    # update the chat response
                    chat_response += answer
                    # display the chunk on the terminal
                    if print_on_terminal:
                        # word wrap
                        if answer in ("```", "``"):
                            block_start = not block_start
                            if block_start:
                                self.word_wrap = False
                            else:
                                self.word_wrap = word_wrap
                        if self.word_wrap:
                            if "\n" in answer:
                                lines = answer.split("\n")
                                for index, line in enumerate(lines):
                                    is_last_line = (len(lines) - index == 1)
                                    self.wrapStreamWords(line, terminal_width)
                                    if not is_last_line:
                                        print("\n", end='', flush=True)
                                        self.line_width = 0
                            else:
                                self.wrapStreamWords(answer, terminal_width)
                        else:
                            print(answer, end='', flush=True) # Print the response
            else:
                finishOutputs(word_wrap, chat_response)
                return None

        finishOutputs(word_wrap, chat_response)



from ..utils.online import get_local_ip
from ..utils.schema import getParameterSchema
from typing import Optional
from tqdm import tqdm
from ollama import Client, Options
from ollama._types import ChatResponse
from ollama import pull
from ollama import list as ollama_ls
import ollama
import re, json

class OllamaLLM:

    DEFAULT_ENDPOINT = f"http://{get_local_ip()}:11434"
    DEFAULT_MODEL = "llama3.2"
    DEFAULT_TEMPERATURE = 0.3
    DEFAULT_MAX_TOKENS = -1
    DEFAULT_CONTEXT_WINDOW = 2048
    DEFAULT_BATCH_SIZE = 512
    DEFAULT_KEEP_ALIVE = "5m"

    @staticmethod
    def getChatCompletion(
            messages: list,
            model: Optional[str]=None,
            model_keep_alive: Optional[str]=None,
            schema: Optional[dict]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            context_window: Optional[int]=None, # applicable to ollama only
            batch_size: Optional[int]=None, # applicable to ollama only
            prefill: Optional[str]=None,
            stop: Optional[list]=None,
            stream: Optional[bool]=False,
            #api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            **kwargs,
    ) -> ChatResponse:
        if prefill:
            messages.append({'role': 'assistant', 'content': prefill})
        model = model if model else OllamaLLM.DEFAULT_MODEL
        # download model if it is not in the model list
        OllamaLLM.downloadModel(model)
        return Client(host=api_endpoint if api_endpoint else OllamaLLM.DEFAULT_ENDPOINT).chat(
            keep_alive=model_keep_alive if model_keep_alive else OllamaLLM.DEFAULT_KEEP_ALIVE,
            model=model,
            messages=messages,
            format=getParameterSchema(schema) if schema else None,
            stream=stream,
            options=Options(
                temperature=temperature if temperature is not None else OllamaLLM.DEFAULT_TEMPERATURE,
                num_ctx=context_window if context_window is not None else OllamaLLM.DEFAULT_CONTEXT_WINDOW,
                num_batch=batch_size if batch_size is not None else OllamaLLM.DEFAULT_BATCH_SIZE,
                num_predict=max_tokens if max_tokens else OllamaLLM.DEFAULT_MAX_TOKENS,
                stop=stop,
                **kwargs,
            ),
        )

    @staticmethod
    def getDictionaryOutput(
            messages: list,
            schema: dict,
            model: Optional[str]=None,
            model_keep_alive: Optional[str]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            context_window: Optional[int]=None, # applicable to ollama only
            batch_size: Optional[int]=None, # applicable to ollama only
            prefill: Optional[str]=None,
            stop: Optional[list]=None,
            #api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            **kwargs,
    ) -> dict:
        completion = OllamaLLM.getChatCompletion(
            messages,
            model=model,
            schema=schema,
            temperature=temperature,
            max_tokens=max_tokens,
            context_window=context_window,
            batch_size=batch_size,
            prefill=prefill,
            stop=stop,
            api_endpoint=api_endpoint,
            model_keep_alive=model_keep_alive,
            **kwargs
        )
        jsonOutput = completion.message.content
        jsonOutput = re.sub("^[^{]*?({.*?})[^}]*?$", r"\1", jsonOutput)
        return json.loads(jsonOutput)

    @staticmethod
    def downloadModel(model: str, force: bool=False) -> bool:
        if not ":" in model:
            model = f"{model}:latest"
        if force or not model in [i.model for i in ollama_ls().models]:
            print(f"Downloading model '{model}' ...")
            try:
                #https://github.com/ollama/ollama-python/blob/main/examples/pull-progress/main.py
                current_digest, bars = '', {}
                for progress in pull(model, stream=True):
                    digest = progress.get('digest', '')
                    if digest != current_digest and current_digest in bars:
                        bars[current_digest].close()

                    if not digest:
                        print(progress.get('status'))
                        continue

                    if digest not in bars and (total := progress.get('total')):
                        bars[digest] = tqdm(total=total, desc=f'pulling {digest[7:19]}', unit='B', unit_scale=True)

                    if completed := progress.get('completed'):
                        bars[digest].update(completed - bars[digest].n)

                    current_digest = digest
            except ollama.ResponseError as e:
                print('Error:', e.error)
                return False
        return True


try:
    from google.genai.types import Content, GenerateContentConfig, SafetySetting, Tool
    from google.genai import Client
except:
    # Google GenAI SDK is not supported on Android Termux
    pass
from typing import Optional, Any
import json, os


class GenaiLLM:

    DEFAULT_API_KEY = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
    DEFAULT_API_PROJECT_ID = ""
    DEFAULT_API_SERVICE_LOCATION = "us-central1"
    DEFAULT_MODEL = "gemini-1.5-pro"
    DEFAULT_TEMPERATURE = 0.3
    DEFAULT_MAX_TOKENS = 8192 # https://cloud.google.com/vertex-ai/generative-ai/docs/learn/models

    @staticmethod
    def toGenAIMessages(messages: dict=[]) -> Optional[list]:
        system_message = ""
        last_user_message = ""
        if messages:
            history = []
            for i in messages:
                role = i.get("role", "")
                content = i.get("content", "")
                if role in ("user", "assistant"):
                    history.append(Content(role="user" if role == "user" else "model", parts=[types.Part.from_text(content)]))
                    if role == "user":
                        last_user_message = content
                elif role == "system":
                    system_message = content
            # remove the last user message
            if history and history[-1].role == "user":
                history = history[:-1]
            else:
                last_user_message = ""
            if not history:
                history = None
        else:
            history = None
        return history, system_message, last_user_message

    @staticmethod
    def getChatCompletion(
            messages: list,
            model: Optional[str]=None,
            schema: Optional[dict]=None,
            temperature: Optional[float]=None,
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            stream: Optional[bool]=False,
            api_key: Optional[str]=None, # enter credentials json file path if using Vertex AI; or enter Google AI API key for accessing Google AI services
            #api_endpoint: Optional[str]=None,
            api_project_id: Optional[str]=None, # applicable to Vertex AI only
            api_service_location: Optional[str]=None, # applicable to Vertex AI only
            **kwargs,
    ) -> Any:
        if not api_key and not GenaiLLM.DEFAULT_API_KEY:
            raise ValueError("API key is required.")
        #if prefill:
        #    messages.append({'role': 'assistant', 'content': prefill})
        # convert messages to GenAI format
        history, system_message, last_user_message = GenaiLLM.toGenAIMessages(messages=messages)
        # create GenAI client
        api_project_id = api_project_id if api_project_id else GenaiLLM.DEFAULT_API_PROJECT_ID
        api_service_location = api_service_location if api_service_location else GenaiLLM.DEFAULT_API_SERVICE_LOCATION
        api_key = api_key if api_key else GenaiLLM.DEFAULT_API_KEY
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = api_key if os.path.isfile(api_key) else ""
        genai_client = Client(vertexai=True, project=api_project_id, location=api_service_location) if os.path.isfile(api_key) and api_project_id and api_service_location else Client(api_key=api_key)
        # format GenAI tool
        if schema:
            name, description, parameters = schema["name"], schema["description"], schema["parameters"]
            if "type" in parameters:
                parameters["type"] = parameters["type"].upper() # Input should be 'TYPE_UNSPECIFIED', 'STRING', 'NUMBER', 'INTEGER', 'BOOLEAN', 'ARRAY' or 'OBJECT' [type=literal_error, input_value='object', input_type=str]
            for key, value in parameters["properties"].items():
                if "type" in value:
                    parameters["properties"][key]["type"] = parameters["properties"][key]["type"].upper() # Input should be 'TYPE_UNSPECIFIED', 'STRING', 'NUMBER', 'INTEGER', 'BOOLEAN', 'ARRAY' or 'OBJECT' [type=literal_error, input_value='object', input_type=str]
            # declare a function
            function_declaration = dict(
                name=name,
                description=description,
                parameters=parameters,
            )
            tool = Tool(
                function_declarations=[function_declaration],
            )
            tools = [tool]
        else:
            tools = None
        # generate content
        genai_config = GenerateContentConfig(
            system_instruction=system_message,
            temperature=temperature if temperature is not None else GenaiLLM.DEFAULT_TEMPERATURE,
            #top_p=0.95,
            #top_k=20,
            candidate_count=1,
            #seed=5,
            max_output_tokens=max_tokens if max_tokens else GenaiLLM.DEFAULT_MAX_TOKENS,
            stop_sequences=stop if stop else ["STOP!"],
            #presence_penalty=0.0,
            #frequency_penalty=0.0,
            safety_settings= [
                SafetySetting(
                    category='HARM_CATEGORY_CIVIC_INTEGRITY',
                    threshold='BLOCK_ONLY_HIGH',
                ),
                SafetySetting(
                    category='HARM_CATEGORY_DANGEROUS_CONTENT',
                    threshold='BLOCK_ONLY_HIGH',
                ),
                SafetySetting(
                    category='HARM_CATEGORY_HATE_SPEECH',
                    threshold='BLOCK_ONLY_HIGH',
                ),
                SafetySetting(
                    category='HARM_CATEGORY_HARASSMENT',
                    threshold='BLOCK_ONLY_HIGH',
                ),
                SafetySetting(
                    category='HARM_CATEGORY_SEXUALLY_EXPLICIT',
                    threshold='BLOCK_ONLY_HIGH',
                ),
            ],
            tools=tools,
        )
        genai_chat = genai_client.chats.create(
            model=model if model else GenaiLLM.DEFAULT_MODEL,
            config=genai_config,
            history=history,
            **kwargs
        )
        return genai_chat.send_message_stream(last_user_message) if stream else genai_chat.send_message(last_user_message)

    @staticmethod
    def getDictionaryOutput(
            messages: list,
            schema: dict,
            model: Optional[str]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            #api_endpoint: Optional[str]=None,
            api_project_id: Optional[str]=None, # applicable to Vertex AI only
            api_service_location: Optional[str]=None, # applicable to Vertex AI only
            **kwargs,
    ) -> dict:
        completion = GenaiLLM.getChatCompletion(
            messages,
            model=model,
            schema=schema,
            temperature=temperature,
            max_tokens=max_tokens,
            #prefill=prefill,
            stop=stop,
            api_key=api_key,
            #api_endpoint=api_endpoint,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            **kwargs
        )
        textOutput = completion.candidates[0].content.parts[0].text
        if textOutput and textOutput.startswith("```json\n"):
            textOutput = textOutput[8:-4]
        return json.loads(textOutput)

class VertexaiLLM:

    @staticmethod
    def getChatCompletion(
            messages: list,
            model: Optional[str]=None,
            schema: Optional[dict]=None,
            temperature: Optional[float]=None,
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            stream: Optional[bool]=False,
            api_key: Optional[str]=None, # enter credentials json file path if using Vertex AI; or enter Google AI API key for accessing Google AI services
            #api_endpoint: Optional[str]=None,
            api_project_id: Optional[str]=None, # applicable to Vertex AI only
            api_service_location: Optional[str]=None, # applicable to Vertex AI only
            **kwargs,
    ) -> Any:
        return GenaiLLM.getChatCompletion(
            messages,
            model=model,
            schema=schema,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            stream=stream,
            api_key=api_key,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            **kwargs
        )

    @staticmethod
    def getDictionaryOutput(
            messages: list,
            schema: dict,
            model: Optional[str]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            #api_endpoint: Optional[str]=None,
            api_project_id: Optional[str]=None, # applicable to Vertex AI only
            api_service_location: Optional[str]=None, # applicable to Vertex AI only
            **kwargs,
    ) -> dict:
        return GenaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            **kwargs
        )



import socket, webbrowser, shutil
from toolmate_sdk.utils.system import runSystemCommand

def isServerAlive(ip, port):
    if ip.lower() == "localhost":
        ip = "127.0.0.1"
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(2)  # Timeout in case of server not responding
    try:
        sock.connect((ip, port))
        sock.close()
        return True
    except socket.error:
        return False

def get_local_ip():
    """
    Gets the local IP address of the machine.
    Returns:
        str: The local IP address.
    """
    try:
        # Create a socket object
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # Connect to a known external server (e.g., Google's DNS server)
        s.connect(("8.8.8.8", 80))
        # Get the local IP address assigned to the socket
        ip_address = s.getsockname()[0]
        s.close()
        return ip_address
    except Exception as e:
        #print(f"Error getting local IP address: {e}")
        return "127.0.0.1"

def openURL(url):
    if shutil.which("termux-open-url"):
        command = f'''termux-open-url "{url}"'''
        runSystemCommand(command)
    else:
        webbrowser.open(url)


from openai import OpenAI
from openai.types.chat import ChatCompletion
from typing import Optional
import json, os


class OpenaiCompatibleLLM:

    DEFAULT_API_KEY = os.getenv("CUSTOM_API_KEY")
    DEFAULT_API_ENDPOINT = ""
    DEFAULT_MODEL = ""
    DEFAULT_TEMPERATURE = 0.3
    DEFAULT_MAX_TOKENS = 4000 # https://docs.github.com/en/github-models/prototyping-with-ai-models#rate-limits

    @staticmethod
    def getChatCompletion(
            messages: list,
            model: Optional[str]=None,
            schema: Optional[dict]=None,
            temperature: Optional[float]=None,
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            stream: Optional[bool]=False,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            api_timeout: Optional[float]=None,
            **kwargs,
    ) -> ChatCompletion:
        #if not api_key and not OpenaiCompatibleLLM.DEFAULT_API_KEY:
        #    raise ValueError("API key is required.")
        #if not api_endpoint and not OpenaiCompatibleLLM.DEFAULT_API_ENDPOINT:
        #    raise ValueError("API endpoint is required.")
        #if prefill:
        #    messages.append({'role': 'assistant', 'content': prefill})
        return OpenAI(api_key=api_key if api_key else OpenaiCompatibleLLM.DEFAULT_API_KEY, base_url=api_endpoint if api_endpoint else OpenaiCompatibleLLM.DEFAULT_API_ENDPOINT).chat.completions.create(
            model=model if model else OpenaiCompatibleLLM.DEFAULT_MODEL,
            messages=messages,
            temperature=temperature if temperature is not None else OpenaiCompatibleLLM.DEFAULT_TEMPERATURE,
            max_tokens=max_tokens if max_tokens else OpenaiCompatibleLLM.DEFAULT_MAX_TOKENS,
            tools=[{"type": "function", "function": schema}] if schema else None,
            tool_choice={"type": "function", "function": {"name": schema["name"]}} if schema else None,
            stream=stream,
            stop=stop,
            timeout=api_timeout,
            **kwargs
        )

    @staticmethod
    def getDictionaryOutput(
            messages: list,
            schema: dict,
            model: Optional[str]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            api_timeout: Optional[float]=None,
            **kwargs,
    ) -> dict:
        completion = OpenaiCompatibleLLM.getChatCompletion(
            messages,
            model=model,
            schema=schema,
            temperature=temperature,
            max_tokens=max_tokens,
            #prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
        return json.loads(completion.choices[0].message.tool_calls[0].function.arguments)



def getRagPrompt(query: str, context: str) -> str:
    return context if not query else f"""# Provided Context

{context}

# My question:

{query}

# Instruction

Select all the relevant information from the provided context to answer my question in as much detail as possible."""


from toolmate_sdk.utils.handle_python_code import fineTunePythonCode
from toolmate_sdk.utils.system import get_linux_distro
import platform

TOOL_PLATFORM = "Linux (" + get_linux_distro().get("name", "") + ")" if platform.system() == "Linux" else platform.system()
TOOL_PLATFORM = TOOL_PLATFORM.replace("()", "")

TOOL_SYSTEM = f"""You are a senior python engineer. Your expertise is to resolve my request, by generating python code that works on {TOOL_PLATFORM}."""

TOOL_SCHEMA = {
    "name": "task",
    "description": "Execute computing task or gain access to device information",
    "parameters": {
        "type": "object",
        "properties": {
            "code": {
                "type": "string",
                "description": "Generate Python code that integrates any relevant packages to resolve my request",
            },
            "title": {
                "type": "string",
                "description": "Title for the task",
            },
            "risk": {
                "type": "string",
                "description": "Assess the risk level of damaging my device upon executing the task. e.g. file deletions or similar significant impacts are regarded as 'high' level.",
                "enum": ["high", "medium", "low"],
            },
        },
        "required": ["code", "title", "risk"],
    },
}

def task(code: str, title: str, risk: str):
    print(f"Running task: {title}")
    #print(f"Risk: {risk}") # implement risk management later
    print()

    refined_python_code = fineTunePythonCode(code)
    exec(refined_python_code, globals())

    print()
    print("Done!")
    return ""

TOOL_METHOD = task


from mistralai import Mistral, ChatCompletionResponse, UNSET, CompletionEvent
from mistralai.utils.eventstreaming import EventStream
from typing import Optional, Union
import json, os


class MistralLLM:
    # docs: https://docs.mistral.ai/

    DEFAULT_API_KEY = os.getenv("MISTRAL_API_KEY")
    DEFAULT_MODEL = "mistral-large-latest"
    DEFAULT_TEMPERATURE = 0.3
    DEFAULT_MAX_TOKENS = 8000 # https://docs.mistral.ai/getting-started/models/models_overview/

    @staticmethod
    def getChatCompletion(
            messages: list,
            model: Optional[str]=None,
            schema: Optional[dict]=None,
            temperature: Optional[float]=None,
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            prefill: Optional[str]=None,
            stop: Optional[list]=None,
            stream: Optional[bool]=False,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            #api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            api_timeout: Optional[int]=None,
            **kwargs,
    ) -> Union[EventStream[CompletionEvent], ChatCompletionResponse]:
        if not api_key and not MistralLLM.DEFAULT_API_KEY:
            raise ValueError("API key is required.")
        if prefill:
            messages.append({'role': 'assistant', 'content': prefill, "prefix": True})
        return Mistral(api_key=api_key if api_key else MistralLLM.DEFAULT_API_KEY).chat.stream(
            model=model if model else MistralLLM.DEFAULT_MODEL,
            messages=messages,
            temperature=temperature if temperature is not None else MistralLLM.DEFAULT_TEMPERATURE,
            max_tokens=max_tokens if max_tokens else MistralLLM.DEFAULT_MAX_TOKENS,
            stop=stop,
            timeout_ms=api_timeout,
            **kwargs
        ) if stream else Mistral(api_key=api_key if api_key else MistralLLM.DEFAULT_API_KEY).chat.complete(
            model=model if model else MistralLLM.DEFAULT_MODEL,
            messages=messages,
            temperature=temperature if temperature is not None else MistralLLM.DEFAULT_TEMPERATURE,
            max_tokens=max_tokens if max_tokens else MistralLLM.DEFAULT_MAX_TOKENS,
            tools=[{"type": "function", "function": schema}] if schema else UNSET,
            tool_choice="any" if schema else None,
            stop=stop,
            timeout_ms=api_timeout,
            **kwargs
        )

    @staticmethod
    def getDictionaryOutput(
            messages: list,
            schema: dict,
            model: Optional[str]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            prefill: Optional[str]=None,
            stop: Optional[list]=None,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            #api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            api_timeout: Optional[int]=None,
            **kwargs,
    ) -> dict:
        completion = MistralLLM.getChatCompletion(
            messages,
            model=model,
            schema=schema,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
        return json.loads(completion.choices[0].message.tool_calls[0].function.arguments)



from toolmate_sdk.utils.manage_package import installPipPackage
REQUIREMENTS = ["googlesearch-python"]
try:
    import googlesearch
except:
    for i in REQUIREMENTS:
        installPipPackage(i)
    import googlesearch

import os, json

TOOL_SYSTEM = "You expertise is to identify keywords for online searches, in order to resolve user request."

TOOL_SCHEMA = {
    "name": "search_google",
    "description": "Search Google for real-time information or latest updates when LLM lacks information",
    "parameters": {
        "type": "object",
        "properties": {
            "keywords": {
                "type": "string",
                "description": "Keywords for online searches",
            },
        },
        "required": ["keywords"],
    },
}

def search_google(keywords):
    info = {}
    for index, item in enumerate(googlesearch.search(keywords, advanced=True, num_results=os.getenv("MAXIMUM_INTERNET_SEARCHES") if os.getenv("MAXIMUM_INTERNET_SEARCHES") else 5)):
        info[f"information {index}"] = {
            "title": item.title,
            "url": item.url,
            "description": item.description,
        }
    return json.dumps(info)

TOOL_METHOD = search_google


from setuptools import setup
import os, shutil

package = "toolmate_sdk"

# update package readme
latest_readme = "README.md" # github repository readme
package_readme = os.path.join(package, "README.md") # package readme
shutil.copy(latest_readme, package_readme)
with open(package_readme, "r", encoding="utf-8") as fileObj:
    long_description = fileObj.read()

# get required packages
install_requires = []
with open(os.path.join(package, "requirements.txt"), "r") as fileObj:
    for line in fileObj.readlines():
        mod = line.strip()
        if mod and not mod.startswith("#"):
            install_requires.append(mod)

# make sure config.py is empty
#open(os.path.join(package, "config.py"), "w").close()

# https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/
setup(
    name="toolmate-sdk",
    version="0.0.11",
    python_requires=">=3.8, <3.13",
    description="ToolMate-SDK: a software developement kit for developing agentic AI applications that support 13 LLM backends and integrate tools and agents. (Developer: Eliran Wong)",
    long_description=long_description,
    author="Eliran Wong",
    author_email="support@toolmate.ai",
    packages=[
        package,
        f"{package}.backends",
        f"{package}.contexts",
        f"{package}.plugins",
        f"{package}.prompts",
        f"{package}.systems",
        f"{package}.tools",
        f"{package}.utils",
    ],
    package_data={
        package: ["*.*"],
        f"{package}.backends": ["*.*"],
        f"{package}.contexts": ["*.*"],
        f"{package}.plugins": ["*.*"],
        f"{package}.prompts": ["*.*"],
        f"{package}.systems": ["*.*"],
        f"{package}.tools": ["*.*"],
        f"{package}.utils": ["*.*"],
    },
    license="GNU General Public License (GPL)",
    install_requires=install_requires,
    extras_require={
        'genai': ["google-genai>=0.6.0"],  # Dependencies for running Vertex AI
    },
    entry_points={
        "console_scripts": [
            #f"tmsdk={package}.main:main", # cli for quick run
        ],
    },
    keywords="toolmate ai sdk anthropic azure chatgpt deepseek genai github googleai groq llamacpp mistral ollama openai vertexai xai",
    url="https://github.com/eliranwong/toolmate-sdk",
    project_urls={
        "Source": "https://github.com/eliranwong/toolmate-sdk",
        "Tracker": "https://github.com/eliranwong/toolmate-sdk/issues",
        "Documentation": "https://github.com/eliranwong/toolmate-sdk/wiki",
        "Funding": "https://www.paypal.me/toolmate",
    },
    classifiers=[
        # Reference: https://pypi.org/classifiers/

        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: End Users/Desktop',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



import os, re, subprocess

def runSystemCommand(command):
    result = subprocess.run(command, shell=True, capture_output=True, text=True)
    output = result.stdout  # Captured standard output
    error = result.stderr  # Captured standard error
    response = ""
    response += f"# Output:\n{output}"
    if error.strip():
        response += f"\n# Error:\n{error}"
    return response

def get_linux_distro():
    """
    Detects the Linux distribution using various methods.

    Returns:
    A dictionary containing information about the distribution, or None if 
    the distribution could not be determined.
    """

    if os.path.isdir("/data/data/com.termux/files/home") and not os.getcwd().startswith("/root"):
        return {"name": "Android Termux", "version": ""}

    # Method 1: Check /etc/os-release (most reliable)
    try:
        with open("/etc/os-release", "r") as f:
            os_release_content = f.readlines()
            distro_info = {}
            for line in os_release_content:
                if "=" in line:
                    key, value = line.strip().split("=", 1)
                    distro_info[key.lower()] = value.strip('"')
        
        # Prioritize 'pretty_name' or 'name' if available.
        if distro_info.get("pretty_name"):
            distro_info["name"] = distro_info["pretty_name"]
        else:
            distro_info["name"] = "Unknown"

        return distro_info
    except FileNotFoundError:
        pass

    # Method 2: Check for distro-specific files in /etc
    distro_files = {
        "redhat": "/etc/redhat-release",
        "fedora": "/etc/fedora-release",
        "centos": "/etc/centos-release",
        "oracle": "/etc/oracle-release",
        "debian": "/etc/debian_version",
        "ubuntu": "/etc/lsb-release",
        "linuxmint": "/etc/lsb-release",
        "gentoo": "/etc/gentoo-release",
        "alpine": "/etc/alpine-release",
        "arch": "/etc/arch-release",
        "manjaro": "/etc/lsb-release",
        "opensuse": "/etc/SuSE-release", 
        "suse": "/etc/SuSE-release" # older SUSE
    }

    for distro_name, file_path in distro_files.items():
        try:
            with open(file_path, "r") as f:
                content = f.read().strip()
                
                if distro_name == "ubuntu" or distro_name == "linuxmint" or distro_name == "manjaro":
                    # Parse lsb-release for Ubuntu, Mint and Manjaro
                    distro_info = {}
                    for line in content.splitlines():
                        key, value = line.split("=")
                        distro_info[key.lower()] = value.strip()
                    if distro_info.get("distrib_description"):
                        return {"name": distro_info["distrib_description"], "version": distro_info.get("distrib_release"), "codename": distro_info.get("distrib_codename")}
                    else:
                        return {"name": distro_info.get("distrib_id"), "version": distro_info.get("distrib_release"), "codename": distro_info.get("distrib_codename")}

                elif distro_name == "debian":
                    # Debian only has a version number 
                    return {"name": "Debian", "version": content}

                elif distro_name == "arch":
                    # Arch Linux typically has an empty /etc/arch-release
                    return {"name": "Arch Linux"}
                    
                else:
                    # Extract name and version for other distributions
                    # This regex tries to handle different formats
                    match = re.search(r"([\w\s]+)[\s|-]*release\s*([\d.abrc]+)?", content, re.IGNORECASE)

                    if match:
                        name = match.group(1).strip()
                        version = match.group(2).strip() if match.group(2) else None
                        
                        # Special case handling for CentOS to differentiate from RHEL
                        if name.lower().startswith("centos"):
                            return {"name": "CentOS", "version": version}
                        else:
                            return {"name": name, "version": version}
                    else:
                        return {"name": distro_name, "version": content}

        except FileNotFoundError:
            pass

    # Method 3: Use platform.linux_distribution() (deprecated)
    # try:
    #   # This is deprecated in Python 3.8+ and removed in Python 3.10
    #   distro_name, version, codename = platform.linux_distribution()
    #   if distro_name:
    #     return {"name": distro_name, "version": version, "codename": codename}
    # except AttributeError:
    #   pass
        
    # Method 4: Use lsb_release command (if available)
    try:
        # Check if lsb_release command exists
        if os.system("which lsb_release > /dev/null 2>&1") == 0:
            distro_info = {}
            output = os.popen("lsb_release -a").read()
            for line in output.splitlines():
                if ":" in line:
                    key, value = line.split(":", 1)
                    distro_info[key.strip().lower().replace(" ", "_")] = value.strip()
            if distro_info.get("distributor_id"):
                return {"name": distro_info["distributor_id"], "version": distro_info.get("release"), "codename": distro_info.get("codename"), "description": distro_info.get("description")}
    except Exception:
        pass

    return {"name": "", "version": ""}  # Could not determine distro


import textwrap, re

def fineTunePythonCode(code):
    # dedent
    code = textwrap.dedent(code).rstrip()
    code = re.sub("^python[ ]*\n", "", code)
    # extract from code block, if any
    if code_only := re.search('```python\n(.+?)```', code, re.DOTALL):
        code = code_only.group(1).strip()
    # make sure it is run as main program
    if "\nif __name__ == '__main__':\n" in code:
        code, main = code.split("\nif __name__ == '__main__':\n", 1)
        code = code.strip()
        main = "\n" + textwrap.dedent(main)
    elif '\nif __name__ == "__main__":\n' in code:
        code, main = code.split('\nif __name__ == "__main__":\n', 1)
        code = code.strip()
        main = "\n" + textwrap.dedent(main)
    else:
        main = ""
    return f"{code}{main}"


from anthropic import Anthropic, NOT_GIVEN
from anthropic.types import Message
from typing import Optional
import os


class AnthropicLLM:
    # docs: https://docs.anthropic.com/en/home

    DEFAULT_API_KEY = os.getenv("ANTHROPIC_API_KEY")
    DEFAULT_MODEL = "claude-3-5-sonnet-latest"
    DEFAULT_TEMPERATURE = 0.3
    DEFAULT_MAX_TOKENS = 8192 # https://docs.mistral.ai/getting-started/models/models_overview/

    @staticmethod
    def getChatCompletion(
            messages: list,
            model: Optional[str]=None,
            schema: Optional[dict]=None,
            temperature: Optional[float]=None,
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            stream: Optional[bool]=False,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            #api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            api_timeout: Optional[float]=None,
            **kwargs,
    ) -> Message:
        if not api_key and not AnthropicLLM.DEFAULT_API_KEY:
            raise ValueError("API key is required.")
        #if prefill:
        #    messages.append({'role': 'assistant', 'content': prefill})
        if schema:
            schema["input_schema"] = schema.pop("parameters")
        return Anthropic(api_key=api_key if api_key else AnthropicLLM.DEFAULT_API_KEY).messages.create(
            model=model if model else AnthropicLLM.DEFAULT_MODEL,
            messages=messages,
            temperature=temperature if temperature is not None else AnthropicLLM.DEFAULT_TEMPERATURE,
            max_tokens=max_tokens if max_tokens else AnthropicLLM.DEFAULT_MAX_TOKENS,
            tools=[schema] if schema else NOT_GIVEN,
            tool_choice={"type": "tool", "name": schema["name"]} if schema else NOT_GIVEN,
            stream=stream,
            stop_sequences=stop,
            timeout=api_timeout,
            **kwargs
        )

    @staticmethod
    def getDictionaryOutput(
            messages: list,
            schema: dict,
            model: Optional[str]=None,
            temperature: Optional[float]=None, 
            max_tokens: Optional[int]=None,
            #context_window: Optional[int]=None, # applicable to ollama only
            #batch_size: Optional[int]=None, # applicable to ollama only
            #prefill: Optional[str]=None,
            stop: Optional[list]=None,
            api_key: Optional[str]=None, # api key for backends that require one; enter credentials json file path if using Vertex AI
            #api_endpoint: Optional[str]=None,
            #api_project_id: Optional[str]=None, # applicable to Vertex AI only
            #api_service_location: Optional[str]=None, # applicable to Vertex AI only
            api_timeout: Optional[float]=None,
            **kwargs,
    ) -> dict:
        completion = AnthropicLLM.getChatCompletion(
            messages,
            model=model,
            schema=schema,
            temperature=temperature,
            max_tokens=max_tokens,
            #prefill=prefill,
            stop=stop,
            api_key=api_key,
            #api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
        outputDictionary = {}
        for i in completion.content:
            if hasattr(i, "input"):
                outputDictionary = i.input
                break
        return outputDictionary
